//---------------------------------------------------------------------------------------------------------------------------
//	Module:	Register
//	Desc:	This is a simple variable width register with enable and synchrnous reset.  Note that in this version enable
//		gates reset.
//---------------------------------------------------------------------------------------------------------------------------
module Register(In, Out, Clock, Enable, Reset);
	parameter		width = 8;
	input	[width-1:0]	In;
	output	[width-1:0]	Out;
	input			Clock, Enable, Reset;

	reg	[width-1:0]	Out;

	always @ (posedge Clock) begin
		if (Enable) begin
			if (Reset) Out <= 0;
			else Out <= In;
		end
	end
endmodule

//---------------------------------------------------------------------------------------------------------------------------
//	Module:	ClippingAdd
//	Desc:	This module is a clipping adder, it will add any two luma or chroma values, clipping them appropriately to
//		ensure that the integrity of the outgoing data stream is maintained
//---------------------------------------------------------------------------------------------------------------------------
module ClippingAdd(Out, In0, In1);
	//-------------------------------------------------------------------------------------------------------------------
	//	I/O
	//-------------------------------------------------------------------------------------------------------------------
	input	[7:0]	In0, In1;
	output	[7:0]	Out;
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Wires
	//-------------------------------------------------------------------------------------------------------------------
	wire	[8:0]	Sum;
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Logic
	//-------------------------------------------------------------------------------------------------------------------
	// Add the two input values normally
	assign Sum = In0 + In1;
	// Clip the output to the 0x10 to 0xEF range
        assign Out = ((Sum > 9'hef) ? 9'hef :
                      ((Sum < 9'h10) ? 9'h10 : Sum));
	//-------------------------------------------------------------------------------------------------------------------
endmodule

//---------------------------------------------------------------------------------------------------------------------------
//	Module:	AlphaMultiply
//	Desc:	Multiply an 8-bit alpha value by a 3-bit intensity value.  This module performs the main composition
//		operation nessecary for variable alpha blending
//---------------------------------------------------------------------------------------------------------------------------
module AlphaMultiply(Intensity, AlphaIn, AlphaOut);
	//-------------------------------------------------------------------------------------------------------------------
	//	Inputs
	//-------------------------------------------------------------------------------------------------------------------
	input	[2:0]	Intensity;
	input	[7:0]	AlphaIn;
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Output
	//-------------------------------------------------------------------------------------------------------------------
	output	[7:0]	AlphaOut;
	reg	[7:0]	AlphaOut;
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Very efficient combinational multiplier
	//-------------------------------------------------------------------------------------------------------------------
	always @ (Intensity or AlphaIn) begin
		casex(Intensity)
			3'b1xx: AlphaOut = AlphaIn;			// Maximum intensity
			3'b011: AlphaOut = AlphaIn[7:1] + AlphaIn[7:2];	// 75% intensity
			3'b010: AlphaOut = AlphaIn[7:1];		// 50% intensity
			3'b001: AlphaOut = AlphaIn[7:2];		// 25% intensity
			3'b000: AlphaOut = 8'h00;			// Minimum intensity
		endcase
	end
	//-------------------------------------------------------------------------------------------------------------------
endmodule

//---------------------------------------------------------------------------------------------------------------------------
//	Module:	VectorBlend
//	Desc:	Blend a vector with the given direction and intensity into the video stream.  Row and Col refere to the pixel
//		row and column within the vector corresponding to the current pixel value on VideoIn.
//---------------------------------------------------------------------------------------------------------------------------
module VectorBlend(Direction, Intensity, Row, Col, VideoIn, VideoOut, Clock, Blanking);
	//-------------------------------------------------------------------------------------------------------------------
	//	Vector Input
	//	These values describe the direction and intensity of the desired vector
	//-------------------------------------------------------------------------------------------------------------------
	input	[3:0]	Direction;
	input	[2:0]	Intensity;
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Control Input
	//	These inputs tell the VectorBlend module which row and column of the vector to use at the current location
	//	as well as disabling all vector blending during the blanking interval
	//-------------------------------------------------------------------------------------------------------------------
	input	[3:0]	Row;		// Which row of the 8x8 vector is on VideoIn right now?
	input	[2:0]	Col;		// Which pixel pair column of the 8x8 vector in on VideoIn right now?
	input		Blanking;	// Disable the vector blending to avoid destroying blanking data
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Raw Video and Clock
	//-------------------------------------------------------------------------------------------------------------------
	input	[7:0]	VideoIn;	// Raw video in, from the output of the decoder
	output	[7:0]	VideoOut;	// Mixed video out to the ADV7194
	input		Clock;		// VE_CLOCK_
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Wires and registers to pipeline this unit to operate at 27MHz
	//	Without these the critical path would be more than 37ns long
	//-------------------------------------------------------------------------------------------------------------------
	wire	[7:0]	SyncVideoIn, SyncVideoIn_, SyncVideoOut;
	wire		SyncBlank_, SyncBlank;
	Register #(9)	VIReg1(.Out({SyncVideoIn_, SyncBlank_}), .In({VideoIn, Blanking}), .Clock(Clock), .Enable(1'b1), .Reset(1'b0));
	Register #(9)	VIReg2(.Out({SyncVideoIn, SyncBlank}), .In({SyncVideoIn_, SyncBlank_}), .Clock(Clock), .Enable(1'b1), .Reset(1'b0));
	Register #(8)	VOReg(.Out(VideoOut), .In(SyncVideoOut), .Clock(Clock), .Enable(1'b1), .Reset(1'b0));
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Load the component (Y,Cr,Cb) value for the current direction, intensity, row and col
	//-------------------------------------------------------------------------------------------------------------------
	wire	[7:0]	MemC, MemL0, MemL1;	// L0, L1 and C alpha factors loaded from memory
	wire	[10:0]	MemAddress;		// Address to load things from
	reg	[7:0]	MemAlpha;		// Alpha Value loaded from memory

	reg	[15:0]	MixValue;

	wire	[1:0]	NComp;			// Next Component
	reg	[1:0]	CComp;			// Current Component

	wire	[7:0]	Component, Alpha;
	wire	[7:0]	Component_, Alpha_;

	// The memory address is a composite of the direction, row and column
	assign	MemAddress =	{Direction[3:2], Row, Direction[1:0], Col};

	// Luma Alpha and Chroma Alpha memories (synchronous memory)
	vectorluma	LumaMem(.addr(MemAddress), .clk(Clock), .dout({MemL1, MemL0}));
	vectorchroma	ChromaMem(.addr(MemAddress), .clk(Clock), .dout(MemC));

	// FSM State encodings
	parameter	COMP_CHROMAB =	2'h0,
			COMP_LUMA0 =	2'h1,
			COMP_CHROMAR =	2'h2,
			COMP_LUMA1 =	2'h3;

	assign	NComp =		CComp + 1;	// Move to the next component
	always @ (posedge Clock) begin		// Simple FSM to determine which component we're currently mixing
		if (Blanking) CComp <= COMP_LUMA1;
		else CComp <= NComp;
	end

	always @ (CComp or MemC or MemL0 or MemL1) begin
		case (CComp)
			COMP_CHROMAB: MemAlpha = MemC;
			COMP_CHROMAR: MemAlpha = MemC;
			COMP_LUMA0: MemAlpha = MemL0;
			COMP_LUMA1: MemAlpha = MemL1;
		endcase
	end

	// Alpha multiplier multiplies an intensity by the alpha value loaded from the memories
	AlphaMultiply	AM(.Intensity(Intensity), .AlphaIn(MemAlpha), .AlphaOut(Alpha_));

	// Pipeline register
	Register #(8)	AReg(.Out(Alpha), .In(Alpha_), .Clock(Clock), .Enable(1'b1), .Reset(1'b0));

	always @ (CComp or Alpha) begin
		case (CComp)
			COMP_CHROMAB: MixValue = 8'hF0 * Alpha;
			COMP_CHROMAR: MixValue = 8'h70 * Alpha;
			COMP_LUMA0: MixValue = 8'h28 * Alpha;
			COMP_LUMA1: MixValue = 8'h28 * Alpha;
		endcase
	end

	// Use only the high 8 bits of the mix value
	assign	Component_ =	MixValue[15:8];

	// Pipeline register
	Register #(8)	CReg(.Out(Component), .In(Component_), .Clock(Clock), .Enable(1'b1), .Reset(1'b0));
	//-------------------------------------------------------------------------------------------------------------------

	//-------------------------------------------------------------------------------------------------------------------
	//	Final mixing
	//-------------------------------------------------------------------------------------------------------------------
	wire	[3:0]	SVAlpha;	// Inverse of Alpha
	wire	[7:0]	Added;		// Result of clipping add
	reg	[7:0]	SyncVideoMixed;	// Scaled version of the input video

	// Invert the Alpha signal from the above block
	assign	SVAlpha =	{(&(~Alpha[7:5])), ~(Alpha[7:5])};

	// Perform 9 level alpha scaling on the input video signal
	always @ (SVAlpha or SyncVideoIn) begin
		casex(SVAlpha)
			4'b1xxx: SyncVideoMixed = SyncVideoIn;
			4'b0111: SyncVideoMixed = SyncVideoIn[7:1] + SyncVideoIn[7:2] + SyncVideoIn[7:3];
			4'b0110: SyncVideoMixed = SyncVideoIn[7:1] + SyncVideoIn[7:2];
			4'b0101: SyncVideoMixed = SyncVideoIn[7:1] + SyncVideoIn[7:3];
			4'b0100: SyncVideoMixed = SyncVideoIn[7:1];
			4'b0011: SyncVideoMixed = SyncVideoIn[7:2] + SyncVideoIn[7:3];
			4'b0010: SyncVideoMixed = SyncVideoIn[7:2];
			4'b0001: SyncVideoMixed = SyncVideoIn[7:3];
			4'b0000: SyncVideoMixed = 8'h00;
		endcase
	end

	// Add the scaled input video to the scaled vector from the above block
	ClippingAdd	CA(.Out(Added), .In0(SyncVideoMixed), .In1(Component));

	// Mux the output between the mixed signal and the raw input signal so that during blanking we dont mess up the video
	assign	SyncVideoOut =	SyncBlank ? SyncVideoIn : Added;
	//-------------------------------------------------------------------------------------------------------------------

endmodule